"""NyxPy public API.

Async SDK for Nyx (bots / user-bots).
"""
from .types import (
    NyxError,
    NyxHTTPError,
    NyxAuthError,
    AuthResult,
    Event,
    EventType,
    Server,
    UserInfo, FriendUser, FriendRequest, Channel, Role,  # NEW
)
from .utils import configure_logging, get_logger
from .http_api import NyxHTTPClient
from .gateway import NyxClient, nyx_event  # NyxClient extends NyxHTTPClient with WS/events

__all__ = [
    # logging
    "configure_logging",
    "get_logger",
    # errors/types
    "NyxError",
    "NyxHTTPError",
    "NyxAuthError",
    "AuthResult",
    "Event",
    "EventType",
    "Server",
    "UserInfo",
    "FriendUser",
    "FriendRequest",
    "Channel",
    "Role",
    # clients
    "NyxHTTPClient",
    "NyxClient",
    # decorators
    "nyx_event",
]

__version__ = "0.1.1"


from __future__ import annotations

import json
from typing import Any, Dict, List, Mapping, Optional

from .http import NyxHTTPClient as _Base
from .types import Server, UserInfo, Member, Event, FriendRequest, Role, Channel
from .http_parsers import (
    parse_user_info, parse_member, parse_friend_request, parse_role, parse_channel,
)

class NyxHTTPClient(_Base):
    """High-level HTTP API on top of the core transport/auth client."""

    # ---- messages ----
    async def send_message(
        self,
        channel_name: str,
        content: str,
        *,
        reply_id: Optional[str] = None,
        reply: Optional[Mapping[str, Any]] = None,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {"channel_name": channel_name, "content": content}
        if reply is not None:
            payload["reply"] = dict(reply)
        elif reply_id:
            payload["reply"] = {"id": str(reply_id)}
        resp = await self.request("POST", "/i/messages/send", json_body=payload)
        try:
            return resp.json()
        except json.JSONDecodeError:
            return {}

    async def reply(self, ev: Event, content: str) -> Dict[str, Any]:
        channel_name: Optional[str] = None
        if isinstance(ev.data, dict):
            cn = ev.data.get("channel_name")
            if isinstance(cn, str) and cn:
                channel_name = cn
        if not channel_name and isinstance(ev.channel, str) and ev.channel:
            channel_name = ev.channel
        if not channel_name:
            raise RuntimeError("Cannot determine channel_name from event to reply().")
        reply_id: Optional[str] = None
        if isinstance(ev.data, dict):
            mid = ev.data.get("id")
            if isinstance(mid, str) and mid:
                reply_id = mid
        if not reply_id:
            raise RuntimeError("Event has no message id to reply to.")
        return await self.send_message(channel_name, content, reply_id=reply_id)

    # ---- friends ----
    async def accept_friend_request(self, request_id: int) -> Dict[str, Any]:
        resp = await self.request("POST", f"/i/friends/accept/{int(request_id)}")
        try:
            return resp.json()
        except json.JSONDecodeError:
            return {}

    async def get_received_friend_requests(self) -> list[FriendRequest]:
        resp = await self.request("GET", "/i/friends/received-requests")
        try:
            data = resp.json()
        except json.JSONDecodeError:
            data = []
        out: list[FriendRequest] = []
        if isinstance(data, list):
            for item in data:
                if isinstance(item, dict):
                    out.append(parse_friend_request(item))
        return out

    # ---- servers/users ----
    async def get_servers(self) -> List[Server]:
        resp = await self.request("GET", "/i/servers/")
        try:
            data = resp.json()
        except json.JSONDecodeError:
            data = []
        servers: List[Server] = []
        if isinstance(data, list):
            for item in data:
                if not isinstance(item, dict):
                    continue
                servers.append(
                    Server(
                        id=int(item.get("id")),
                        name=str(item.get("name", "")),
                        owner_id=int(item.get("owner_id")),
                        avatar_url=str(item.get("avatar_url", "")),
                    )
                )
        self._server_list = servers
        return servers

    async def get_user_info(self, *, cached: bool = True, force_refresh: bool = False) -> UserInfo:
        if (cached and not force_refresh) and self.current_user is not None:
            return self.current_user
        resp = await self.request("GET", "/i/user/me")
        try:
            data = resp.json()
        except json.JSONDecodeError:
            data = {}
        user = parse_user_info(data)
        self.current_user = user
        return user

    async def get_server_members(self, server_id: int) -> list[Member]:
        resp = await self.request("GET", f"/i/servers/{server_id}/members")
        try:
            data = resp.json()
        except json.JSONDecodeError:
            data = []
        members: list[Member] = []
        if isinstance(data, list):
            for item in data:
                if not isinstance(item, dict):
                    continue
                m = parse_member(item)
                members.append(m)
                self._member_by_user[m.id] = m
        self._server_members[server_id] = {m.id for m in members}
        return members

    async def ensure_server_members_cached(self, server_id: int) -> list[Member]:
        if server_id not in self._server_members:
            return await self.get_server_members(server_id)
        ids = self._server_members[server_id]
        return [self._member_by_user[uid] for uid in ids if uid in self._member_by_user]

    async def get_member(self, server_id: int, user_id: int, *, fetch_if_missing: bool = True) -> Optional[Member]:
        uid = int(user_id)
        if server_id in self._server_members and uid in self._server_members[server_id]:
            return self._member_by_user.get(uid)
        if not fetch_if_missing:
            return self._member_by_user.get(uid)
        await self.get_server_members(server_id)
        return self._member_by_user.get(uid)

    # Channels
    async def get_channels(self, server_id: int) -> list[Channel]:
        """GET /i/channels/?server_id=..."""
        resp = await self.request("GET", "/i/channels/", params={"server_id": int(server_id)})
        try:
            data = resp.json()
        except json.JSONDecodeError:
            data = []
        out: list[Channel] = []
        if isinstance(data, list):
            for item in data:
                if isinstance(item, dict):
                    out.append(parse_channel(item))
        return out

    # Roles (all in server)
    async def get_roles(self, server_id: int) -> list[Role]:
        """GET /i/roles/?server_id=..."""
        resp = await self.request("GET", "/i/roles/", params={"server_id": int(server_id)})
        try:
            data = resp.json()
        except json.JSONDecodeError:
            data = []
        out: list[Role] = []
        if isinstance(data, list):
            for item in data:
                if isinstance(item, dict):
                    out.append(parse_role(item))
        return out

    # Roles for current user in server
    async def get_my_roles(self, server_id: int) -> list[Role]:
        """GET /i/roles/user?server_id=..."""
        resp = await self.request("GET", "/i/roles/user", params={"server_id": int(server_id)})
        try:
            data = resp.json()
        except json.JSONDecodeError:
            data = []
        out: list[Role] = []
        if isinstance(data, list):
            for item in data:
                if isinstance(item, dict):
                    out.append(parse_role(item))
        return out



from __future__ import annotations

import logging
from typing import Any, Dict, Mapping

_LOGGER_NAME = "nyxpy"


def get_logger() -> logging.Logger:
    logger = logging.getLogger(_LOGGER_NAME)
    if not any(isinstance(h, logging.NullHandler) for h in logger.handlers):
        logger.addHandler(logging.NullHandler())
    return logger


def configure_logging(level: int = logging.INFO) -> None:
    logger = get_logger()
    logger.handlers = [h for h in logger.handlers if not isinstance(h, logging.NullHandler)]
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(asctime)s | %(levelname)s | %(name)s | %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(level)


def _safe_json(obj: Mapping[str, Any]) -> Dict[str, Any]:
    REPLACEMENTS = {"password", "pass", "token", "access_token", "refresh_token"}
    redacted: Dict[str, Any] = {}
    for k, v in obj.items():
        redacted[k] = "***" if k in REPLACEMENTS else v
    return redacted


def _safe_json_obj(obj: Mapping[str, Any]) -> Dict[str, Any]:
    try:
        return _safe_json(obj)
    except Exception:
        return {"_raw": str(obj)}



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Mapping, Optional, List


class NyxError(Exception):
    pass


class NyxHTTPError(NyxError):
    def __init__(self, status_code: int, message: str, *, payload: Optional[Mapping[str, Any]] = None):
        super().__init__(f"HTTP {status_code}: {message}")
        self.status_code = status_code
        self.payload = payload


class NyxAuthError(NyxHTTPError):
    pass


@dataclass(slots=True)
class AuthResult:
    access_token: Optional[str]
    refresh_token: Optional[str]
    token_type: str = "Bearer"
    raw: Dict[str, Any] = None


class EventType(str, Enum):
    NEW_MESSAGE = "NEW_MESSAGE"
    TYPING = "TYPING"
    PRESENCE_UPDATE = "PRESENCE_UPDATE"
    CONNECT = "CONNECT"
    DISCONNECT = "DISCONNECT"
    RAW = "RAW"
    FRIEND_REQUEST = "FRIEND_REQUEST"


@dataclass(slots=True)
class Event:
    type: EventType
    data: Dict[str, Any]
    channel: Optional[str] = None
    raw: Optional[Dict[str, Any]] = None


# HTTP models
@dataclass(slots=True)
class Server:
    id: int
    name: str
    owner_id: int
    avatar_url: str


@dataclass(slots=True)
class UserInfo:
    id: int
    email: str
    username: str
    avatar_url: str

@dataclass(slots=True)
class Member:
    id: int
    email: str
    username: str
    avatar_url: str
    roles: List[int]

@dataclass(slots=True)
class FriendUser:
    id: int
    email: str
    username: str
    avatar_url: str

@dataclass(slots=True)
class FriendRequest:
    id: int
    sender_id: int
    receiver_id: int
    status: str
    sender: Optional[FriendUser] = None

@dataclass(slots=True)
class Channel:
    id: int
    name: str
    server_id: int
    type: str
    position: int
    parent_channel_id: Optional[int]
    roles_overrides: List[Any]
    users_overrides: List[Any]

@dataclass(slots=True)
class Role:
    id: int
    name: str
    color: str
    server_id: int
    position: int
    display_role_members: bool
    permissions: List[str]


from __future__ import annotations

import asyncio
import json
from typing import Any, Dict, List, Mapping, Optional

import httpx

from .types import AuthResult, NyxAuthError, NyxError, NyxHTTPError, Server, UserInfo, Member, Event, FriendRequest, \
    FriendUser
from .utils import _safe_json, _safe_json_obj, get_logger
from .http_parsers import parse_user_info, parse_member, parse_friend_user, parse_friend_request


class NyxHTTPClient:
    BASE_URL = "https://nyx-app.ru/api/v1"

    def __init__(
        self,
        *,
        base_url: str | httpx.URL | None = None,
        timeout: float = 15.0,
        headers: Optional[Mapping[str, str]] = None,
        transport: Optional[httpx.AsyncBaseTransport] = None,
        user_agent: str = "NyxPy/0.1 (+https://github.com/kolya5544/nyxpy)",
        logger=None,
        verify: httpx._types.VerifyTypes | None = None,
        trust_env: bool = True,
    ) -> None:
        self._logger = logger or get_logger()
        self._base_url = httpx.URL(str(base_url or self.BASE_URL))
        default_headers: Dict[str, str] = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": user_agent,
        }
        if headers:
            default_headers.update(headers)

        self._client = httpx.AsyncClient(
            base_url=self._base_url,
            timeout=timeout,
            headers=default_headers,
            transport=transport,
            http2=True,
            verify=verify,
            trust_env=trust_env,
        )
        self._access_token: Optional[str] = None
        self._refresh_token: Optional[str] = None
        self._token_type: str = "Bearer"
        self.current_user: Optional[UserInfo] = None
        self._server_list: List[Server] = []
        self._member_by_user: dict[int, Member] = {}  # user_id -> Member
        self._server_members: dict[int, set[int]] = {}  # server_id -> {user_ids}

    # ---- lifecycle ----
    async def aclose(self) -> None:
        await self._client.aclose()

    async def __aenter__(self) -> "NyxHTTPClient":  # pragma: no cover
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:  # pragma: no cover
        await self.aclose()

    # ---- auth ----
    async def login(self, email: str, password: str) -> AuthResult:
        """POST /auth/login and store access_token for all future requests."""
        path = "/auth/login"
        payload = {"email": email, "password": password}
        self._logger.debug("POST %s payload=%s", path, _safe_json(payload))
        try:
            resp = await self._client.post(path, json=payload)
        except httpx.HTTPError as e:
            raise NyxError(f"Network error during login: {e}") from e

        if resp.status_code >= 400:
            message = self._extract_error_message(resp)
            raise NyxAuthError(resp.status_code, message, payload=_safe_json_obj(payload))

        data = self._json_or_empty(resp)
        result = self._parse_auth_response(data)
        self._apply_auth(result)  # sets Authorization for ALL subsequent requests
        if hasattr(self, "get_user_info"):
            try:
                await self.get_user_info()  # provided by high-level API subclass
            except Exception:
                pass
        self._logger.info("Authenticated as %s", email)
        return result

    def _parse_auth_response(self, data: Mapping[str, Any]) -> AuthResult:
        candidates = ("access_token", "accessToken", "token", "jwt", "session")
        access_token = None
        for k in candidates:
            v = data.get(k)
            if isinstance(v, str) and v:
                access_token = v
                break
        if not access_token and isinstance(data.get("data"), Mapping):
            nested = data["data"]
            for k in candidates:
                v = nested.get(k)
                if isinstance(v, str) and v:
                    access_token = v
                    break

        refresh_token = (
            data.get("refresh_token")
            or data.get("refreshToken")
            or (data.get("data", {}).get("refresh_token") if isinstance(data.get("data"), Mapping) else None)
        )
        token_type = data.get("token_type") or data.get("tokenType") or data.get("type") or "Bearer"

        return AuthResult(
            access_token=access_token,
            refresh_token=refresh_token if isinstance(refresh_token, str) else None,
            token_type=str(token_type) if token_type else "Bearer",
            raw=dict(data),
        )

    def _apply_auth(self, auth: AuthResult) -> None:
        if not auth.access_token:
            raise NyxAuthError(200, "Login succeeded but no token was found in response", payload=auth.raw)
        self._access_token = auth.access_token
        self._token_type = auth.token_type or "Bearer"
        self._refresh_token = auth.refresh_token
        self._client.headers["Authorization"] = f"{self._token_type} {self._access_token}"
        self.current_user = None

    @property
    def is_authenticated(self) -> bool:
        return bool(self._access_token)

    # ---- generic request helper ----
    async def request(
        self,
        method: str,
        path: str,
        *,
        json_body: Optional[Mapping[str, Any]] = None,
        params: Optional[Mapping[str, Any]] = None,
        headers: Optional[Mapping[str, str]] = None,
        retry: int = 1,
        retry_backoff: float = 0.5,
    ) -> httpx.Response:
        if headers:
            merged_headers = {**self._client.headers, **headers}
        else:
            merged_headers = self._client.headers

        attempt = 0
        while True:
            try:
                resp = await self._client.request(
                    method,
                    path,
                    json=json_body,
                    params=params,
                    headers=merged_headers,
                )
            except httpx.HTTPError as e:
                self._logger.warning("Network error %r on %s %s", e, method, path)
                if attempt >= retry:
                    raise NyxError(f"Network error: {e}") from e
            else:
                if resp.status_code in (429,) or 500 <= resp.status_code < 600:
                    if attempt < retry:
                        delay = retry_backoff * (2**attempt)
                        self._logger.debug("Retrying %s %s in %.2fs (status=%s)", method, path, delay, resp.status_code)
                        await asyncio.sleep(delay)
                    else:
                        self._raise_for_status(resp)
                        return resp
                else:
                    self._raise_for_status(resp)
                    return resp
            attempt += 1

    # ---- local helpers (HTTP-only) ----
    @staticmethod
    def _json_or_empty(resp: httpx.Response) -> Dict[str, Any]:
        try:
            return resp.json()
        except json.JSONDecodeError:
            return {}

    @staticmethod
    def _extract_error_message(resp: httpx.Response) -> str:
        try:
            data = resp.json()
            return str(
                data.get("message") or data.get("error") or data.get("detail") or data.get("msg") or data
            )
        except json.JSONDecodeError:
            return resp.text or resp.reason_phrase

    @staticmethod
    def _raise_for_status(resp: httpx.Response) -> None:
        if 400 <= resp.status_code:
            msg = NyxHTTPClient._extract_error_message(resp)
            if resp.status_code in (401, 403):
                raise NyxAuthError(resp.status_code, msg)
            raise NyxHTTPError(resp.status_code, msg)



from __future__ import annotations

import asyncio
import contextlib
import inspect
import json
import logging
import ssl
from collections import defaultdict, Counter
from dataclasses import asdict
from typing import Any, Dict, Mapping, Optional, Set, Iterable

import websockets
from websockets import exceptions as wse

from .http_api import NyxHTTPClient
from .types import Event, EventType, NyxAuthError, NyxError, Server
from .utils import get_logger


class _Dispatcher:
    def __init__(self) -> None:
        self._handlers: Dict[EventType, list] = defaultdict(list)

    def register(self, event_type: EventType, func):
        if inspect.iscoroutinefunction(func):
            coro = func
        else:
            async def coro(ev: Event):  # type: ignore[misc]
                return func(ev)
        self._handlers[event_type].append(coro)
        return func

    async def dispatch(self, ev: Event) -> None:
        for handler in list(self._handlers.get(ev.type, [])):
            try:
                await handler(ev)
            except Exception as e:  # pragma: no cover
                get_logger().exception("Handler error on %s: %r", ev.type, e)
        if ev.type is not EventType.RAW:
            for handler in list(self._handlers.get(EventType.RAW, [])):
                try:
                    await handler(ev)
                except Exception as e:  # pragma: no cover
                    get_logger().exception("Handler error on RAW: %r", e)


_dispatcher = _Dispatcher()


class _NoxEventDecorator:
    NEW_MESSAGE = EventType.NEW_MESSAGE
    TYPING = EventType.TYPING
    PRESENCE_UPDATE = EventType.PRESENCE_UPDATE
    CONNECT = EventType.CONNECT
    DISCONNECT = EventType.DISCONNECT
    RAW = EventType.RAW
    FRIEND_REQUEST = EventType.FRIEND_REQUEST

    def __call__(self, *, type: EventType):
        def deco(func):
            return _dispatcher.register(type, func)
        return deco


nyx_event = _NoxEventDecorator()


class NyxClient(NyxHTTPClient):
    """HTTP + WebSocket client with decorator-based events."""

    def __init__(self):
        super().__init__()
        self._ws_ready = asyncio.Event()

    # ---- WebSocket lifecycle ----
    async def connect_ws(
        self,
        token: Optional[str] = None,
        *,
        url: str = "wss://nyx-app.ru/api/v2/i/ws",
        reconnect: bool = True,
        ping_interval: float = 25.0,
        ssl_context: ssl.SSLContext | bool | None = ssl.SSLContext(),
        subprotocols: Optional[Iterable[str]] = None,
        extra_headers: Optional[Mapping[str, str]] = None,
        client_name: str = "js",
    ) -> None:
        self._ws_ready.clear()
        self._ws_token = token or self._access_token
        if not self._ws_token:
            raise NyxAuthError(401, "No token for WS. Call login() first or pass token=.")
        self._ws_url = f"{url}?token={self._ws_token}"
        self._ws_reconnect = reconnect
        self._subscriptions: Set[str] = set()
        self._ws_ssl = ssl_context
        self._ws_subprotocols = list(subprotocols) if subprotocols else None
        self._ws_headers = dict(extra_headers) if extra_headers else None
        self._ws_client_name = client_name
        self._ws_task = asyncio.create_task(self._ws_loop(self._ws_url, ping_interval))
        try:
            await asyncio.wait_for(self._ws_ready.wait(), timeout=None)
        except Exception:
            # if the loop already crashed, raise its exception
            if self._ws_task.done():
                self._ws_task.result()
            raise

    async def close_ws(self) -> None:
        task = getattr(self, "_ws_task", None)
        if task:
            task.cancel()
            with contextlib.suppress(Exception):
                await task
        ws = getattr(self, "_ws", None)
        if ws:
            with contextlib.suppress(Exception):
                await ws.close()

    async def _ws_loop(self, url: str, ping_interval: float) -> None:
        backoff = 1.0
        while True:
            try:
                self._logger.info("WS connecting: %s", url)
                async with websockets.connect(
                        url,
                        ssl=self._ws_ssl,
                        ping_interval=ping_interval,
                        ping_timeout=20,
                        close_timeout=5,
                        max_size=8 * 1024 * 1024,
                        open_timeout=15,
                        subprotocols=self._ws_subprotocols or ["json"],
                        additional_headers=self._ws_headers,
                ) as ws:
                    self._ws = ws
                    self._ws_id = 0
                    self._ka_task = asyncio.create_task(self._keepalive(ws, interval=10))
                    # send CONNECT like the browser does (include name; many servers also want token here)
                    await self._ws_send({"connect": {"name": self._ws_client_name}})
                    # get user data
                    userData = await self.get_user_info()
                    id = userData.id
                    await self._ws_send({"subscribe":{"channel":f"user:{id}","data":{"user_id":id}}})
                    self._ws_ready.set()
                    await _dispatcher.dispatch(Event(EventType.CONNECT, data={}, raw={"url": url}))
                    async for message in ws:
                        obj = json.loads(message)
                        if message == "{}":
                            await self._ws_send_raw("{}")
                            continue
                        ev = self._map_ws_to_event(obj)
                        if ev.type == EventType.NEW_MESSAGE and isinstance(ev.channel, str) and ev.channel.startswith(
                                "server:"):
                            try:
                                # channel format: server:{server_id}:{...}
                                server_id = int(ev.channel.split(":", 2)[1])

                                # ---- attach sender's member (unchanged) ----
                                sender_id = None
                                if isinstance(ev.data, dict):
                                    sid = ev.data.get("sender_id") or ev.data.get("user_id") or ev.data.get("author_id")
                                    if isinstance(sid, str):
                                        sender_id = int(sid) if sid.isdigit() else None
                                    elif isinstance(sid, (int, float)):
                                        sender_id = int(sid)
                                if sender_id is not None:
                                    member = await self.get_member(server_id, sender_id, fetch_if_missing=True)
                                    if member:
                                        ev.data["member"] = asdict(member)

                                # ---- NEW: attach replied-to member, if present ----
                                # looks inside ev.data["data"]["reply"]["sender_id"]
                                reply_sender_id = None
                                inner = ev.data.get("data") if isinstance(ev.data, dict) else None
                                if isinstance(inner, dict):
                                    rep = inner.get("reply")
                                    if isinstance(rep, dict):
                                        rsid = rep.get("sender_id")
                                        if isinstance(rsid, str):
                                            reply_sender_id = int(rsid) if rsid.isdigit() else None
                                        elif isinstance(rsid, (int, float)):
                                            reply_sender_id = int(rsid)

                                if reply_sender_id:
                                    rmember = await self.get_member(server_id, reply_sender_id, fetch_if_missing=True)
                                    if rmember:
                                        # prefer to attach alongside the nested reply payload
                                        if isinstance(inner, dict):
                                            inner["reply_member"] = asdict(rmember)
                                        elif isinstance(ev.data, dict):
                                            ev.data["reply_member"] = asdict(rmember)
                            except Exception as e:  # pragma: no cover
                                self._logger.debug("Failed to enrich NEW_MESSAGE for %s: %r", ev.channel, e)

                        await _dispatcher.dispatch(ev)
            except asyncio.CancelledError:
                self._logger.info("WS loop cancelled")
                raise
            except wse.InvalidStatusCode as e:
                self._logger.warning("WS handshake failed with HTTP %s", getattr(e, "status_code", "?"))
            except (wse.ConnectionClosedOK, wse.ConnectionClosedError) as e:
                self._logger.warning("WS closed: code=%s reason=%r", getattr(e, "code", "?"), getattr(e, "reason", ""))
            except ssl.SSLError as e:
                self._logger.warning("WS SSL error: %r", e)
            except Exception as e:  # pragma: no cover
                self._logger.warning("WS error: %r", e)
            finally:
                # stop keepalive if running
                ka = getattr(self, "_ka_task", None)
                if ka:
                    ka.cancel()
                    with contextlib.suppress(Exception):
                        await ka
                await _dispatcher.dispatch(Event(EventType.DISCONNECT, data={}, raw={}))

            if not getattr(self, "_ws_reconnect", True):
                break
            await asyncio.sleep(min(backoff, 30))
            backoff *= 2

    async def _keepalive(self, ws, interval: float) -> None:
        """Send WS control pings periodically (defensive against idle timeouts)."""
        try:
            while True:
                await asyncio.sleep(interval)
                # iterate over all servers, send presence
                for s in self._server_list:
                    await self._ws_send({"presence": {"channel":f"server:{s.id}:general"}})
                await ws.ping()
                #await ws.send("{}")
        except asyncio.CancelledError:
            return

    # ---- Mapping & send helpers ----
    def _map_ws_to_event(self, obj: Dict[str, Any]) -> Event:
        # push
        if isinstance(obj, dict) and "push" in obj:
            push = obj["push"]
            pub = push.get("pub", {}) if isinstance(push, dict) else {}
            chan = pub.get("channel") or push.get("channel")
            data = pub.get("data", {})
            raw_type = data.get("type") if isinstance(data, dict) else None
            payload = None
            if isinstance(data, dict):
                payload = data.get("data")
                if isinstance(payload, str):
                    try:
                        payload = json.loads(payload)
                    except json.JSONDecodeError:
                        pass
            if raw_type == "message":
                return Event(EventType.NEW_MESSAGE, data=payload or {}, channel=chan, raw=obj)
            if raw_type == "typing":
                return Event(EventType.TYPING, data=payload or {}, channel=chan, raw=obj)
            if raw_type == "friend_request":
                return Event(EventType.FRIEND_REQUEST, data=payload or {}, channel=chan, raw=obj)
            return Event(EventType.RAW, data=data if isinstance(data, dict) else {"data": data}, channel=chan, raw=obj)

        # presence
        if isinstance(obj, dict) and "presence" in obj:
            pres = obj["presence"]
            chan = pres.get("channel") if isinstance(pres, dict) else None
            details = pres.get("presence") if isinstance(pres, dict) else pres
            if not isinstance(details, dict):
                details = {"presence": details}
            return Event(EventType.PRESENCE_UPDATE, data=details, channel=chan, raw=obj)

        # subscribe/connect acks fall back to RAW (useful for debugging)
        if isinstance(obj, dict) and ("connect" in obj or "subscribe" in obj):
            return Event(EventType.RAW, data=obj, raw=obj)

        return Event(EventType.RAW, data=obj if isinstance(obj, dict) else {"raw": obj}, raw=obj)

    async def _ws_send(self, payload: Dict[str, Any]) -> None:
        ws = getattr(self, "_ws", None)
        if ws is None:
            raise NyxError("WS not connected; call connect_ws() first")
        self._ws_id = getattr(self, "_ws_id", 0) + 1
        frame = dict(payload)
        frame["id"] = self._ws_id
        jsonstr = json.dumps(frame, separators=(",", ":"))
        #logging.debug(f"[SENT] {jsonstr}")
        await ws.send(jsonstr)


    async def _ws_send_raw(self, text: str) -> None:
        """Send a raw text frame without injecting an 'id' field."""
        ws = getattr(self, "_ws", None)
        if ws is None:
            raise NyxError("WS not connected; call connect_ws() first")
        await ws.send(text)

    # ---- Presence (snapshot) ----
    async def ws_presence(self, channel: str) -> None:
        """Request presence snapshot for a channel."""
        userInfo = await self.get_user_info()
        userId = userInfo.id
        await self._ws_send({"subscribe": {"channel": channel, "data":{"user_id":userId}}})
        self._subscriptions.add(channel)

    # ---- Subscriptions ----
    async def ws_subscribe_channel(self, channel: str, *, data: Optional[Mapping[str, Any]] = None) -> None:
        """Generic subscribe helper (useful for 'user:<id>' or any channel requiring extra data)."""
        payload: Dict[str, Any] = {"subscribe": {"channel": channel}}
        if data:
            payload["subscribe"]["data"] = dict(data)
        await self._ws_send(payload)
        self._subscriptions.add(channel)

    async def ws_subscribe_user(self, user_id: int) -> None:
        """Subscribe to personal user channel per your flow: {"subscribe":{"channel":"user:<id>","data":{"user_id":<id>}}}"""
        await self.ws_subscribe_channel(f"user:{user_id}", data={"user_id": user_id})

    async def ws_subscribe_server(self, server_id: int, *, channel_name: str = "general") -> None:
        """Subscribe to a specific server's channel. Many deployments accept presence as the join mechanism."""
        channel = f"server:{server_id}:{channel_name}"
        # If your backend prefers explicit subscribe for servers too, switch to ws_subscribe_channel(channel)
        await self.ws_presence(channel)

    # replace the method with this version
    async def ws_subscribe_all(self, *, channel_name: str = "general") -> None:
        """Fetch servers via HTTP and subscribe to each one's channel,
        excluding any servers whose IDs are non-unique in the response.
        """
        servers: list[Server] = await self.get_servers()
        counts = Counter(s.id for s in servers)

        # keep only servers that appear exactly once; drop all duplicates entirely
        unique_servers = [s for s in servers if counts[s.id] == 1]

        # (optional) debug log duplicate IDs we skipped
        dup_ids = [sid for sid, c in counts.items() if c > 1]
        if dup_ids:
            self._logger.debug("Skipping duplicate server IDs: %s", dup_ids)

        for s in unique_servers:
            await self.ws_subscribe_server(s.id, channel_name=channel_name)
            await self._ws_send({"presence": {"channel": f"server:{s.id}:general"}})


from __future__ import annotations
from typing import Mapping, Any
from .types import UserInfo, Member, FriendUser, FriendRequest, Role, Channel


def parse_user_info(data: Mapping[str, Any]) -> UserInfo:
    return UserInfo(
        id=int(data.get("id", 0)),
        email=str(data.get("email", "")),
        username=str(data.get("username", "")),
        avatar_url=str(data.get("avatar_url", "")),
    )

def parse_member(data: Mapping[str, Any]) -> Member:
    roles = data.get("roles") or []
    if not isinstance(roles, list):
        roles = []
    return Member(
        id=int(data.get("id", 0)),
        email=str(data.get("email", "")),
        username=str(data.get("username", "")),
        avatar_url=str(data.get("avatar_url", "")),
        roles=[int(r) for r in roles if isinstance(r, (int, str))],
    )

def parse_friend_user(data: Mapping[str, Any]) -> FriendUser:
    return FriendUser(
        id=int(data.get("id", 0)),
        email=str(data.get("email", "")),
        username=str(data.get("username", "")),
        avatar_url=str(data.get("avatar_url", "")),
    )

def parse_friend_request(data: Mapping[str, Any]) -> FriendRequest:
    sender = data.get("sender")
    sender_obj = parse_friend_user(sender) if isinstance(sender, Mapping) else None
    return FriendRequest(
        id=int(data.get("id", 0)),
        sender_id=int(data.get("sender_id", 0)),
        receiver_id=int(data.get("receiver_id", 0)),
        status=str(data.get("status", "")),
        sender=sender_obj,
    )

def parse_channel(data: Mapping[str, Any]) -> Channel:
    parent = data.get("parent_channel_id")
    parent_id = int(parent) if isinstance(parent, (int, str)) and str(parent).isdigit() else None
    return Channel(
        id=int(data.get("id", 0)),
        name=str(data.get("name", "")),
        server_id=int(data.get("server_id", 0)),
        type=str(data.get("type", "")),
        position=int(data.get("position", 0)),
        parent_channel_id=parent_id,
        roles_overrides=list(data.get("roles_overrides", []) or []),
        users_overrides=list(data.get("users_overrides", []) or []),
    )

def parse_role(data: Mapping[str, Any]) -> Role:
    perms = data.get("permissions") or []
    if not isinstance(perms, list):
        perms = []
    return Role(
        id=int(data.get("id", 0)),
        name=str(data.get("name", "")),
        color=str(data.get("color", "")),
        server_id=int(data.get("server_id", 0)),
        position=int(data.get("position", 0)),
        display_role_members=bool(data.get("display_role_members", False)),
        permissions=[str(p) for p in perms],
    )
